

from .chained_transform import ChainedTransform

__all__ = [
    'empty_transform', 'resize_rdcrop_flip', 'resize_centercrop_flip',
]


def empty_transform():
    x = ChainedTransform()
    x = x.ToPILImage().ToTensor()
    return x


def resize_rdcrop_flip(size1, size2, **args):
    """
        Args:
            size1: resize target size.
            size2: rdcrop target size.
    """
    x = ChainedTransform().ToPILImage()
    x = x.Resize(size=size1, **args).RandomCrop(size=size2, **args).RandomHorizontalFlip().ToTensor()
    return x


def resize_centercrop_flip(size1, size2,  **args):
    """
        Args:
            size1: resize target size.
            size2: centercrop target size.
    """
    x = ChainedTransform().ToPILImage()
    x = x.Resize(size=size1, **args).CenterCrop(size=size2, **args).RandomHorizontalFlip().ToTensor()
    return x



import imageio
import re
import matplotlib.pyplot as plt
import shutil
import os
import cv2
import numpy as np
import torch

import sys
sys.path.append('.')


class Visualize():
    def __init__(self, log_fp='rec.log', cache_dir='build/cache'):
        self.log_fp = log_fp
        self.cache_dir = cache_dir
        self.data = []
        self.axis_type = 'plt'
        self.legends = {}
        self.dtype = 'curve'

    @staticmethod
    def __read__(fp, _hash, target=None):
        with open(fp, 'r') as f:
            for line in f.readlines():
                if 'testing#' in line:
                    continue
                if _hash in line:
                    found = Visualize.__re__(line)
                    if not found:
                        continue
                    if not target:
                        yield found
                    if target == found['metrics_name']:
                        yield found

    @staticmethod
    def __re__(line):
        try:
            res = re.search(r'\[(?P<epoch>\d+)\,\ +(?P<iter>\d+)\/(?P<total_iter>\d+)]\ +(?P<metrics_name>.*)\:\ +(?P<value>.*)', line)
            return res.groupdict()
        except Exception:
            return None

    @staticmethod
    def __smooth__(scalar, weight=0.7):
        last = scalar[0]
        smoothed = []
        for point in scalar:
            smoothed_val = last * weight + (1 - weight) * point
            smoothed.append(smoothed_val)
            last = smoothed_val
        return smoothed

    def _read_curve(self, _hash, metrics_name):
        for _dict in self.__read__(fp=self.log_fp, _hash=_hash, target=metrics_name):
            x = float(_dict['epoch'])+float(_dict['iter'])/float(_dict['total_iter'])
            y = float(_dict['value'])
            yield x, y

    def _clean_cache(self):
        """
            clean cache dir.
        """
        if os.path.exists(self.cache_dir):
            shutil.rmtree(self.cache_dir)
        os.mkdir(self.cache_dir)
        print(f'cache dir {self.cache_dir} cleaned.')

    def _check_dtype(self, dtype):
        if self.dtype != dtype and len(self.data) > 0:
            print('push failed, clear stack first.')
            raise NotImplementedError
        self.dtype = dtype

    def _plot_curve(self, axis_type, **kwargs):
        if axis_type == 'plt':
            if 'plt' not in self.legends:
                self.legends['plt'] = []
            for d in self.data:
                data_x = d['data_x']
                data_y = d['data_y']
                description = d['description']
                self.legends.append(description['legend'])
                plt.plot(data_x, data_y, **kwargs)
        elif axis_type == 'twinx':
            if 'acc_ax' not in self.legends:
                self.legends['acc_ax'] = []
                self.legends['loss_ax'] = []
            for d in self.data:
                data_x = d['data_x']
                data_y = d['data_y']
                description = d['description']
                if 'loss' in description['metrics_name']:
                    self.legends['loss_ax'].append(description['legend'])
                    self.loss_ax.plot(data_x, data_y, **kwargs)
                elif 'acc' in description['metrics_name']:
                    self.legends['acc_ax'].append(description['legend'])
                    self.acc_ax.plot(data_x, data_y, **kwargs)

    def import_curve(self, _hash, metrics_name, legend=None, smooth=0):
        self._check_dtype('curve')
        data = [(x, y) for x, y in self._read_curve(_hash, metrics_name)]
        data_x, data_y = zip(*data)
        if legend is None:
            legend = f'{metrics_name}@{_hash}'
        description = {
            'metrics_name': metrics_name,
            '_hash': _hash,
            'legend': legend,
        }
        self.data.append({
            'data_x': data_x,
            'data_y': self.__smooth__(data_y, weight=smooth),
            'description': description,
        })
        return self

    def plot(self, **kwargs):
        self._plot_curve(axis_type=self.axis_type, **kwargs)
        self.data = []
        return self

    def save(self, path):
        """
            save image to path.
            Args:
                path:
                **args(for image only):
        """
        if self.dtype == 'image':
            for image_idx, (d, annotation) in enumerate(self.data):
                savefig(d, f'{path}/{image_idx}.png', size=(d.shape[0], d.shape[1]), annotation=annotation)
            self.data = []
        elif self.dtype == 'curve':
            if self.axis_type == 'plt':
                plt.legend(self.legends['plt'])
            elif self.axis_type == 'twinx':
                self.acc_ax.legend(self.legends['acc_ax'])
                self.loss_ax.legend(self.legends['loss_ax'])
            self.legends = {}
            plt.show()
            plt.savefig(path, pad_inches=0)
            print(f'saved as {path}')
        return self

    def clear(self):
        """
            clear plt canvas.
        """
        self._clean_cache()
        plt.close('all')

    def twinx(self):
        self.axis_type = 'twinx'
        _, acc_ax = plt.subplots()
        loss_ax = acc_ax.twinx()
        acc_ax.set_xlabel("epoches")
        acc_ax.set_ylabel("accuracy")
        loss_ax.set_ylabel("loss")
        self.acc_ax = acc_ax
        self.loss_ax = loss_ax
        return self

    def import_image(self, x, annotation='NA'):
        """
            read tensor as image,
            1-dim tensors are displayed as heatmap, see heatmap().
            Args:
                x: input tensor.
        """
        self._check_dtype('image')
        if x.shape[0] == 1:
            self.data.append((heatmap(x), annotation))
        else:
            self.data.append((tensor2img(x), annotation))
        return self

    # --------------------------------unimplemented below --------------------------------------------

    @staticmethod
    def __rmdupcurve__(x, idx):
        new_x = []
        dup_start = x[idx][0]
        for i, (epoch, data) in enumerate(x):
            if epoch >= dup_start and i < idx:
                continue
            new_x.append((epoch, data))
        return new_x


def savefig(x, path, size=(448, 448), annotation=''):
    fig = plt.gcf()  # generate outputs
    plt.imshow(x, aspect='equal'), plt.axis('off'), fig.set_size_inches(size[1]*12/300.0, size[0]*12/300.0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator()), plt.gca().yaxis.set_major_locator(plt.NullLocator()), plt.subplots_adjust(top=1, bottom=0, left=0, right=1, hspace=0, wspace=0), plt.margins(0, 0)
    plt.text(0, 0, annotation, color='white', size=4, ha="left", va="top", bbox=dict(boxstyle="square", ec='black', fc='black'))
    plt.savefig(path, dpi=300, pad_inches=0)    # visualize masked image


def heatmap(x):
    heatmap = x[0].cpu().numpy()
    heatmap = heatmap/np.max(heatmap)
    # must convert to type unit8
    heatmap = np.uint8(255 * heatmap)
    heatmap = cv2.applyColorMap(heatmap, cv2.COLORMAP_JET)
    return heatmap


def tensor2img(x, imtype=np.uint8):
    """"将tensor的数据类型转成numpy类型，并反归一化.

    Parameters:
        input_image (tensor) --  输入的图像tensor数组
        imtype (type)        --  转换后的numpy的数据类型
    """
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    if not isinstance(x, np.ndarray):
        if isinstance(x, torch.Tensor):  # get the data from a variable
            image_tensor = x.data
        else:
            return x
        image_numpy = image_tensor.cpu().float().numpy()  # convert it into a numpy array
        if image_numpy.shape[0] == 1:  # grayscale to RGB
            image_numpy = np.tile(image_numpy, (3, 1, 1))
        for i in range(len(mean)):
            image_numpy[i] = image_numpy[i] * std[i] + mean[i]
        image_numpy = image_numpy * 255
        image_numpy = np.transpose(image_numpy, (1, 2, 0))  # post-processing: tranpose and scaling
    else:  # if it is a numpy array, do nothing
        image_numpy = x
    return image_numpy.astype(imtype)


def makegif(self, from_dir='build/cache', fp='build/test.gif', sort_func=lambda x: x, reverse=False):
    """
        make gif from images.
        Args:
            from_dir:
            fp:
            sort_func:
    """
    files = [x for x in os.listdir(from_dir) if x.endswith('jpg') or x.endswith('png')]
    gif_images = [imageio.imread(f'{from_dir}/{img_file}') for img_file in files]
    gif_images.sort(sort_func, reverse=reverse)
    imageio.mimsave(fp, gif_images, fps=8)
    return fp



import scipy.io as scio
import torch
import numpy as np

from .utils import Dataset
from labvision.transforms import casnet_fixation_transform


def __maxidx__(_list):
    return _list.index(max(_list))


class EMOd(Dataset):
    label_types = ['single', 'distribution', 'vad',  'eyetrack']

    def __init__(self, root='external/EMOd', train=True, transform=None, mask_transform=None, eyetrack_map=True):
        self.eyetrack_map = eyetrack_map
        self.mask_transform = casnet_fixation_transform() if mask_transform is None else mask_transform
        super().__init__(root, train, transform)

    def __gensplit__(self, root):
        super().__gensplit__(root, 'EMOdImages1019')

    def __readgt__(self, root):
        _dict = {}
        # for train/test split and some inaccessable data from IAPS
        for x in scio.loadmat(f'{root}/allfindata1019_renamed.mat')['allfindata1019']:
            _dict[int(x[0])] = x
        return [_dict[int(x.split('/')[-1].replace('.jpg', ''))] for x in self.data]

    def __loaddata__(self, root, train):
        super().__loaddata__(root, train, image_dir='EMOdImages1019')

    def __loadtarget__(self, root, spliter=' ', start_line=1):
        for index, x in enumerate(self.__readgt__(root)):
            y1 = [float(_str) for _str in x[3:6]]
            y2 = [float(_str) for _str in x[6:15]]
            self.ys['single'].append(__maxidx__(y2))
            self.ys['distribution'].append(y2)
            self.ys['vad'].append(y1)
            self.ys['eyetrack'].append(self.data[index].replace('EMOdImages1019', 'FixationMap/Continous_map'))

    def __totorch__(self):
        for key in self.ys.keys():
            if key != 'eyetrack':
                self.ys[key] = torch.from_numpy(np.array(self.ys[key]))

    def __getitem__(self, index):
        img = self.__cvimg__(self.data[index])
        img, target = super().__getitem__(index)
        _single, _distribution, _vad, _eyetrack = target
        if self.eyetrack_map:
            _eyetrack = self.__cvimg__(_eyetrack)
            _eyetrack = self.mask_transform(_eyetrack)
            _eyetrack = torch.mean(_eyetrack, dim=0)
        return img, [_single, _distribution, _vad, _eyetrack]



import json
import os
import random
import string
import time

import torch
from torch.autograd import Variable


class Status():
    def __init__(self):
        """
            inner status class of AutoControl,
            supports reading /saving /loading.
        """
        self.status_dict = {'hash': self.__genhash__()}

    @staticmethod
    def __genhash__():
        """
            generates a hash with ASCII letters and digits,
            always starts with a letter (for markdown usage).
        """
        _hash_head = ''.join(random.sample(string.ascii_letters, 1))
        _hash_body = ''.join(random.sample(string.ascii_letters+string.digits, 7))
        return _hash_head+_hash_body

    def update(self, status_dict):
        self.status_dict.update(status_dict)

    def epoch(self):
        if 'train' not in self.status_dict:
            return 0
        return self.status_dict['train']['epoch']

    def iter(self):
        return self.status_dict['train']['iter']

    def epoch_finished(self):
        return self.status_dict['train']['epoch_finished']

    def freeze(self, fp):
        with open(fp, 'w') as f:
            json.dump(self.status_dict, f)

    def load(self, fp):
        with open(fp, 'r') as f:
            status_dict = json.load(f)
            self.update(status_dict)

    def hash(self):
        return self.status_dict['hash']

    def optimizer(self):
        return self.status_dict['optimizer']


class AutoControl():

    def __init__(self,
                 dataset,
                 model,
                 dataset_transforms=None, dataset_args={},
                 optimizer=None,
                 loss_func=None, **fpargs):
        """
            automatic train/eval/stop controller for torch.nn.Module,

            Args:
                dataset:
                model:
                dataset_transforms:
                dataset_args:
                optimizer:
                optimizer_args:
                fpargs:
                loss_func:

            example:
                >>> import labvision
                >>> from labvision.io import AutoControl
                >>> import torch.nn.functional as F

                >>> def loss_func(logits, y):
                >>>     x, v = logits
                >>>     return F.CrossEntropyLoss(x, y)+F.CrossEntropyLoss(v, y)

                >>> control = AutoControl(
                >>>     dataset=labvision.datasets.EMOd,
                >>>     dataset_transforms=(labvision.transforms.centercrop_224, labvision.transforms.centercrop_224),
                >>>     model=torchvision.models.resnet50(),
                >>>     optimizer=torch.optim.SGD,
                >>>     optimizer_args={'lr': 0.001, 'weight_decay': 0.9},
                >>>     log_fp='rec.log',
                >>>     loss_func=loss_func,
                >>> )


                >>> for model, status in control.step(interval=5):
                >>>     if status.epoch_finished():
                >>>         metrics = control.eval(model)
                >>>         control.log(metrics)

                >>>     # torch.save(model, f'build/{status.hash}.pth')
                >>>     status.freeze()

        """
        if loss_func:
            self.loss_function = loss_func

        self._init_datasets(dataset, dataset_transforms, **dataset_args)
        self.model = model
        self._init_optimizer(optimizer)
        self.fps = self._check_fpargs(**fpargs)
        self.status = Status()
        self.under_test = False

    def __train__(self, yield_batches, start_epoch=0):
        """
            train over epoches infinitely (not really),
            Args:
                yield_batches: batches to report loss.
                start_epoch: epoch to start from.
        """
        epoch = start_epoch-1
        while True:
            loss = 0.0
            epoch += 1
            for i, sample in enumerate(self.trainloader, 0):
                loss += self.__step__(sample, train=True)
                if (i+1) % yield_batches == 0:
                    loss /= yield_batches
                    epoch_finished = False
                elif i == len(self.trainloader)-1:
                    loss /= len(self.trainloader) % yield_batches
                    epoch_finished = True
                else:
                    continue
                status = {
                    'loss': loss,
                    'epoch': epoch,
                    'iter': i+1,
                    'epoch_finished': epoch_finished,
                }
                loss = 0.0
                yield status

    def __val__(self, yield_batches):
        """
            evaluate eval_loss for fixed batches.
            Args:
                yield_batches: batches used for a single eval.
        """
        while True:
            loss = 0.0
            for i, sample in enumerate(self.valloader, 0):
                loss += self.__step__(sample, train=False)
                if (i+1) % yield_batches == 0:
                    loss /= yield_batches
                    yield {
                        'loss': loss,
                    }
                    loss = 0.0

    def __step__(self, sample, train=True):
        """
            Args:
                sample: batched sample from dataloader.
                train: train=True if step for train.
        """
        x, y = self.__read__(sample)
        if train:
            self.model.train()
            self.optimizer.zero_grad()
            logits = self.forward(x)
            loss = self.loss_function(logits, y)
            loss.backward()
            self.optimizer.step()
        else:
            self.model.eval()
            with torch.no_grad():
                logits = self.forward(x)
                loss = self.loss_function(logits, y)
        return loss.item()

    def __log__(self, msg, display=True):
        """
            Args:
                msg:
                display:
        """
        line = f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}]<{self.status.hash()}>\t'+msg
        if self.under_test:
            line = f'testing# {line}'
        open(self.fps['log_fp'], 'a').write(line+'\n')
        if display:
            print(line)
        return self

    @staticmethod
    def __read__(sample):
        """
            read x,y from sample,
            override if needed.
        """
        x, y = sample
        return Variable(x).cuda(), Variable(y).cuda()

    @staticmethod
    def __eval__(logits, y, ks=(1, 3)):
        """
            Args:
                logits:
                y:
                ks: top-k accs to be evaluated.
        """
        metrics = {f'acc@top{k}': 0 for k in ks}
        for k in ks:
            metrics[f'acc@top{k}'] += (logits.topk(max((1, k)), 1, True, True)[1] == y.view(-1, 1)).sum().float().item()/logits.shape[0]
        return metrics

    def _check_fpargs(self, log_fp='rec.log', **args):
        fp_list = {
            'log_fp': log_fp,
        }
        for key in fp_list:
            print(f':: checking fp: {fp_list[key]}')
            assert os.path.exists(fp_list[key])
        print(':: fp check passed.')
        return fp_list

    def _init_datasets(self, dataset, dataset_transforms, batch_size=32, num_workers=2, **args):
        """
            Args:
                dataset:
                dataset_transforms: transform (for all) or Tuple(transform_train, transform_test)
                batch_size:
                num_workers:

                **args: args for Dataset.__init__().
        """
        if type(dataset_transforms) is tuple:
            _transform_train, _transform_test = dataset_transforms
        else:
            _transform_train = _transform_test = dataset_transforms
        trainset = dataset(train=True, transform=_transform_train, **args)
        testset = dataset(train=False, transform=_transform_test, **args)
        self.trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size, shuffle=True, num_workers=num_workers)
        self.valloader = torch.utils.data.DataLoader(testset, batch_size=batch_size, shuffle=True, num_workers=num_workers)
        self.testloader = torch.utils.data.DataLoader(testset, batch_size=batch_size, shuffle=False, num_workers=num_workers)

    def _init_optimizer(self, optimizer):
        if optimizer is None:
            optimizer = torch.optim.SGD(params=self.model.parameters(), lr=1e-3)
        self.optimizer = optimizer

    def load_status(self, fp, recover_optimizer=False):
        """
            load status from fp,
            Args:
                fp:
                recover_optimizer: recover_optimizer=True if need to recover specific optimizer state_dict.
        """
        self.status.load(fp)
        if recover_optimizer:
            self.optimizer.load_state_dict(self.status.optimizer())
        return self

    def save_status(self, fp):
        """
            save status to fp,
            Args:
                fp:
        """
        self.status.freeze(fp)
        return self

    def save_model(self, fp):
        """
            save model to fp,
            Args:
                fp:
        """
        state_dict = self.model.state_dict()
        torch.save(state_dict, fp)
        return self

    def load_model(self, fp):
        """
            load frozen model from fp,
            Args:
                fp:
        """
        state_dict = torch.load(fp)
        self.model.load_state_dict(state_dict)
        return self

    def log(self, msg, display=True):
        """
            display(optional) & save log.
            Args:
                msg: str or metrics dict.
                display:

            example:
                >>> ac.log('acc@top1:'+acc)
                >>> ac.log(metrics)
        """
        if type(msg) is dict:
            for k in msg:
                self.__log__(f'[{self.status.epoch()}, {self.status.iter():5d}/{len(self.trainloader)}] {k}: {msg[k]}', display=display)
            return self
        return self.__log__(f'[{self.status.epoch()}, {self.status.iter():5d}/{len(self.trainloader)}] {msg}', display=display)

    def check(self):
        """
            check if autocontrol will work properly,
            all loops are cut for 1 loop only in this mode,
            also logs are annotated to be disabled for visualize.
        """
        self.under_test = True
        self.__log__('testing ..')
        self.__log__(self.__str__())
        _ = next(self.step(interval=1, val_interval=1))
        _ = self.eval(auto_log=True)
        self.__log__('test passed.')
        self.under_test = False

    def step(self, interval=1, val_interval=4, auto_log=True):
        """
            step to the next checkpoint,
            works as a generator.

            Args:
                interval:
                val_interval:
                auto_log:
        """

        train_generator = self.__train__(yield_batches=interval, start_epoch=self.status.epoch())
        val_generator = self.__val__(yield_batches=val_interval)

        while True:
            status_train = next(train_generator)
            status_val = next(val_generator)
            status_optimizer = self.optimizer.state_dict()
            status_dict = {'train': status_train, 'val': status_val, 'optimizer': status_optimizer}
            self.status.update(status_dict)
            if auto_log:
                self.log(f'train_loss: {status_train["loss"]}')
                self.log(f'val_loss: {status_val["loss"]}')
            yield self.model, self.status

    def eval(self, auto_log=True):
        """
            evaluate model accuracy,
            Args:
                auto_log:
        """
        self.model.eval()
        metrics = {}
        with torch.no_grad():
            for sample in self.testloader:
                x, y = self.__read__(sample)
                logits = self.forward(x)
                batch_metrics = self.__eval__(logits, y)
                for key in batch_metrics:
                    if key not in metrics:
                        metrics[key] = 0
                    metrics[key] += batch_metrics[key]/len(self.testloader)
                if self.under_test:
                    break
        if auto_log:
            self.log(metrics)
        return metrics

    def forward(self, x):
        """
            Args:
                x:
            override if needed.
        """
        return self.model(x)

    @staticmethod
    def loss_function(logits, y):
        """
            Args:
                logits:
                y:

            override if needed.
        """
        return torch.nn.functional.cross_entropy(logits, y)


if __name__ == "__main__":
    import labvision
    import torchvision

    transforms = (
        labvision.transforms.resize_rdcrop_flip(224, (224, 224)),
        labvision.transforms.resize_centercrop_flip(224, (224, 224)),
    )

    ac = AutoControl(
        labvision.datasets.FI,
        torchvision.models.resnet50(pretrained=True),
        dataset_transforms=transforms,
        loss_func=lambda logits, y: torch.nn.functional.cross_entropy(logits, y)
    )



import json
import os
import random
import sys

import cv2
import numpy as np
import torch
import torch.utils.data as data
from labvision.transforms import empty_transform
sys.path.append('.')


def __maxidx__(_list):
    return _list.index(max(_list))


class DatasetNotFoundException(Exception):
    def __init__(self, fp):
        super().__init__()
        self.fp = fp

    def __str__(self):
        raise NotImplementedError  # TODO: add download src
        return f'Dataset not found in {self.fp}, you can download from ?.'


class Dataset(data.Dataset):
    label_types = ['single', 'distribution']

    def __init__(self, root, train=True, transform=None):
        """
            Basic dataset class,
            index should be like this:
                root_dir
                ->  img_dir
                ->  ground_truth.txt

            Args:
                root:
                train:
                transform:
        """
        self.transform = transform
        self._empty_transform = empty_transform()
        self.__loaddata__(root, train)
        self.ys = {x: [] for x in self.label_types}
        self.__loadtarget__(root)
        self.__totorch__()

    def __gensplit__(self, root, img_dir='images', seed=0):
        """
            generate train/test split as split.json,
            Args:
                root:
                img_dir:
                seed:
        """
        np.random.seed(seed)
        _list = [x for x in os.listdir(f'{root}/{img_dir}')]
        random.shuffle(_list)
        pivod = int(len(_list)*0.2)
        _dict = {'train': _list[pivod:], 'test': _list[:pivod]}
        with open(f'{root}/split.json', 'w') as f:
            json.dump(_dict, f)

    def __loaddata__(self, root, train, img_dir='images'):
        """
            pre-load image path,
            Args:
                root:
                train:
                img_dir:

            override if more complex func needed.
        """
        if not os.path.exists(root):
            raise DatasetNotFoundException(root)
        _path = f'{root}/split.json'
        if not os.path.exists(_path):
            self.__gensplit__(root)
        with open(_path, 'r') as f:
            _json = json.load(f)
            self.data = [f'{root}/{img_dir}/{x}' for x in _json['train' if train else 'test']]

    def __readgt__(self, root, spliter=' ', start_line=1, suffix=''):
        """
            read groundtruth from ground_truth.txt,
            Args:
                root:
                spliter: spliter for labels
                start_line: line to start from
                suffix: image file suffix

            override if more complex func needed.
        """
        _dict = {}
        with open(f'{root}/ground_truth.txt', 'r') as f:
            for line in f.readlines()[start_line:]:  # images-name Amusement Awe Contentment Excitement Anger Disgust Fear Sadness
                data = line.split(spliter)
                _dict[f'{data[0]}{suffix}'] = data
        return [_dict[x.split('/')[-1]] for x in self.data]

    def __loadtarget__(self, **params):
        """
            sort groundtruth into different types,
            'single': one-hot label (maximum),
            'distribution': distributional label (votes),

            Args:
                see __readgt__()

            override if more complex func needed.
        """
        for x in self.__readgt__(**params):
            y = [int(_str.replace('\n', '')) for _str in x[1:]]
            self.ys['single'].append(__maxidx__(y))
            self.ys['distribution'].append(y)

    def __totorch__(self):
        """
            turn targets into torch.Tensor,
        """
        for key in self.ys.keys():
            self.ys[key] = torch.from_numpy(np.array(self.ys[key]))

    def __cvimg__(self, path):
        """
            read image in RGB mode.
        """
        img_cv = cv2.imread(path)
        img_cv = cv2.cvtColor(img_cv, cv2.COLOR_BGR2RGB)
        return img_cv

    def __getitem__(self, index):
        """
            function for torch.util.data.Dataset class,
            returns image, (target_1, target_2, ..., target_n)

            Args:
                index:
        """
        img = self.__cvimg__(self.data[index])
        img = self._empty_transform(img) if self.transform is None else self.transform(img)
        target = [self.ys[k][index] for k in self.ys.keys()]
        return img, target

    def __len__(self):
        """
            function for torch.util.data.Dataset class.
        """
        return len(self.data)
